
#' List valid configuration fields for the hybrid parametrisation
#' mode of CATS.
#'
#' @description
#' The returned data frame contains valid configuration parameters. Each
#' row describes one parameter.
#' The columns are 'field' for the parameter name, 'mode' for the parametrisation
#' mode, and 'required' if the field is mandatory for CATS to run.
#' The columns 'minimum' and 'maximum' are used for range checks (if not NA).
#' The column 'type' specifies the parameter's expected data type, and 'subtype'
#' the type of parameter's list elements, if the data type is 'list'.
#' The names are of the form section.key_name which corresponds to a
#' configuration parameter "key name" for the configuration section "section", as
#' described in the CATS manual.
#'
#' @details
#' The section "suitability" is equivalent to a configuration section of
#' "environment" of type "suitability". Overlays are currently not supported.
#'
#'
#' @return A data frame containing valid configuration field names, and their metadata.
#' @export
#'
get_cats_hybrid_fields <- function() {
    result <- cats_config_fields
    hybrid <- result$mode == "hybrid"
    result <- result[hybrid, ]
    return(result)
}


#' Create a new CATS configuration data frame
#'
#' @param mode character string: CATS parametrisation mode. Currently only 'hybrid' is supported.
#' @param ... configuration parameters. See  [get_cats_hybrid_fields()] for a list of parameters.
#'
#' @description
#' Creates a data frame representing one or more CATS configurations.
#' Each row describes a CATS configuration, each column a configuration parameter.
#' This data frame is initialized with this function, and can either be filled
#' directly, or using the [set_cats_config_row_data()] helper function.
#'
#' @return A data frame with a single row.  The data frame is pre-filled with all required
#' and the most commonly used optional parameters for the hybrid configuration mode, in
#' addition to all other parameters passed to this function.
#' All values are initially set to NA if not otherwise specified.
#'
#' @seealso
#' The data frame created with this function is also used with
#' [set_cats_config_row_data()], [write_cats_config()] and, [run_cats()].
#' For a list of available configuration parameters, see the 'field' column of
#' [get_cats_hybrid_fields()] ## FIXME
#' @export
#'
cats_default_config <- function(mode = "hybrid", ...) {

    if (is.null(mode) || class(mode) != "character") {
        stop("parameter 'mode' must of be of class 'character'")
    }

    if (mode != "hybrid") {
        stop("Currently only mode = \"hybrid\" is supported")
    }

    args_ <- list(...)
    fields <- get_cats_hybrid_fields()
    result <- data.frame()
    required <- fields$mode == "hybrid" & fields$required == T

    for (row in 1:nrow(fields)) {
        if (fields[row, "required"] != T) {
            next
        }
        name <- fields[row, "field"]
        type <- fields[row, "type"]
        subtype <- fields[row, "subtype"]
        if (type == "numeric" || type == "integer" || subtype == "numeric") {
            result[name] <- numeric()
        } else if (type == "character" || (type == "list" && subtype == "character")) {
            result[name] <- character()
        } else if (type == "logical") {
            result[name] <- logical()
        } else {
            result[name] <- c()
            warning(sprintf("unspecified type '%s'", type))
        }
    }
    if (length(args_) > 0) {
        result[1, ] <- set_cats_config_row_data(result[1, ], ...)
    } else {
        result[1, ] <- NA
    }

    comment(result) <- mode

    return(result)
}


#' Add and modify values of a single CATS configuration
#'
#' @param configuration_row a single row data frame
#' @param ... configuration parameters. See  [get_cats_hybrid_fields()] for a list of parameters.
#'
#' @return The data frame passed to this function, with the specified parameters set and valdiated.
#' @export
#'
#' @seealso
#' [cats_default_config()] to create an empty CATS configuration data.frame
set_cats_config_row_data <- function(configuration_row, ...) {
    valid_excerpt(configuration_row)
    args_ <- list(...)
    list_vars <- c("dispersal.kernel_filenames", "dispersal.minimum_weights", "dispersal.maximum_weights")
    valid <- cats_validation_object()
    if (!length(args_)) {
        warning("No data specified")
        return(configuration_row)
    }

    for (n in colnames(configuration_row)) {
        if (is.null(args_[[n]])) {
            next
        }

        if (n %in% list_vars) {
            configuration_row[, n][[1]] <- list(args_[[n]])
        } else if (class(configuration_row[, n]) == class(args_[[n]])) {
            configuration_row[, n] <- args_[n]
        } else {
            stop(sprintf("found argument '%s' with wrong type. Got %s, expected %s.", n, class(args_[[n]]), class(configuration_row[, n])))
        }

        valid <- add_validation(valid, check_configuration_fields(configuration_row, n, args_[[n]]))
    }
    check_validation(valid)

    return(configuration_row)
}

#' Add and modify values of a single CATS configuration
#'
#' @param config A CATS configuration data.frame
#' @param general.run_name character vector. The run name of the configuration to be modified. If
#' the run name is not found, a new row will be appended.
#' @param ... configuration parameters. See  [get_cats_hybrid_fields()] for a list of parameters.
#'
#' @return The modified or appended CATS configuration data.frame
#' @export
set_cats_config_run_information <- function(config, general.run_name, ...)
{
    sel <- which(config$general.run_name == general.run_name)
    excerpt <- config[sel, ]
    if (nrow(excerpt) == 1) {
        config[sel ,] <- set_cats_config_row_data(excerpt, ...)
    } else if (nrow(excerpt) == 0) {
        excerpt <- cats_default_config(mode="hybrid", general.run_name=general.run_name, ...)
        config <- rbind(config, excerpt)
    } else {
        stop(sprintf("Error: %d rows with unique column general.run_name == %s found",
                     nrow(excerpt), general.run_name))
    }

    return (config)
}


keyname <- function(string, category) {
    string <- gsub(sprintf("^%s.", category), "", string)
    string <- gsub("_", " ", string)
    return(string)
}

is_required <- function(field_name) {
    data <- cats_config_fields


    row <- data[data$field == field_name, ]
    if (nrow(row) == 0) {
        return(FALSE)
    }
    required <- row$required


    return(required)
}

key_val_to_string <- function(varname, value) {
    category <- gsub("\\..*", "", varname, perl = T)


    key <- keyname(varname, category)
    if (is.na(value) && is_required(varname) == TRUE) {
        stop(sprintf("required value '%s' is NA", varname))
    }

    known_fields <- cats_config_fields$field

    if (is.logical(value)) {
        if (value) {
            value <- "yes"
        } else {
            value <- "no"
        }
    }

    if (varname %in% known_fields) {
        return(paste(key, " = ", value, sep = ""))
    } else if (startsWith(varname, "overlay_")) {
        return(paste(key, " = ", value, sep = ""))
    } else {
        return(paste(key, " = ", value, " # (unknown, probably ignored)", sep = ""))
    }
}

add_values <- function(catsconfig, category_name, row, ignore = c()) {
    category_name <- sprintf("%s.", category_name)
    result <- character()

    for (name in colnames(catsconfig)) {
        if (name %in% ignore) {
            next
        }
        if (startsWith(name, category_name)) {
            result <- paste(result, key_val_to_string(name, catsconfig[row, name]), sep = "\n")
        }
    }
    return(result)
}


add_list <- function(excerpt, name) {
    category <- gsub("\\..*", "", name, perl = T)

    key <- keyname(name, category)

    values <- excerpt[1, name][[1]]
    values_string <- ""
    for (i in 1:length(values)) {
        values_string <- paste(values_string, trimws(values[i]), sep = "")
        if (i < length(values)) {
            values_string <- paste(values_string, ", ", sep = "")
        }
    }
    result <- sprintf("%s = %s\n", key, values_string)

    return(result)
}


#' Write on or more CATS configuration files
#'
#' Writes one CATS configuration file (in INI format) for each row in th
#' CATS configuration data frame.
#' All relative paths in the configuration file are relative to the working directory.
#' The existence of referenced files is checked.
#'
#' It is not necessary to call this function yourself to run a CATS simulation.
#'
#' @param catsconfig A data frame with a CATS configuration
#' @param working_directory The path to the working directory. This is used
#' to verify that all referenced files are present.
#'
#' @return A vector of CATS configuration files (in INI format).
#' @export
#'
write_cats_config <- function(catsconfig, working_directory) {
    files <- c()
    cats_mode <- "hybrid"
    valid <- cats_validation_object()
    if (!is.null(attributes(catsconfig)$comment)) {
        cats_mode <- attributes(catsconfig)$comment
    }

    for (row in nrow(catsconfig)) {
        overlays <- c()
        for (name in names(catsconfig)) {
            if (startsWith(name, "overlay_")) {
                name <- unlist(strsplit(name, ".", fixed = T))[[1]]
                identifier <- unlist(strsplit(name, "_"))[[2]]
                if (identifier %in% overlays) {

                } else {
                    overlays <- c(overlays, identifier)
                }
            }
        }

        result <- character()

        species_name <- catsconfig[row, "species.name"]

        if (is.na(species_name)) {
            stop("species.name is NA")
        }

        valid <- add_validation(valid, check_cats_configuration(catsconfig))

        result <- paste(result, "[general]", sep = "")
        result <- paste(result, add_values(catsconfig, "general", row), sep = "\n")
        result <- paste(result, "\n", sep = "\n")



        result <- paste(result, sprintf("[species:%s]", species_name), sep = "")
        result <- paste(result, add_values(catsconfig, "species", row, ignore = c("species.name")), sep = "\n")
        result <- paste(result, sprintf("dispersal = %s", species_name), sep = "\n")
        if (cats_mode == "hybrid") {
            result <- paste(result, sprintf("suitability = %s\n", species_name), sep = "\n")
        }
        result <- paste(result, "\n", sep = "\n")

        result <- paste(result, sprintf("[dispersal:%s]", species_name), sep = "")
        result <- paste(result, add_list(catsconfig[row, ], "dispersal.kernel_filenames"), sep = "\n")
        result <- paste(result, add_list(catsconfig[row, ], "dispersal.minimum_weights"), sep = "")
        result <- paste(result, add_list(catsconfig[row, ], "dispersal.maximum_weights"), sep = "")


        if (cats_mode == "hybrid") {
            result <- paste(result, sprintf("[environment:%s]", species_name), sep = "\n")
            result <- paste(result, "\ntype = suitability", sep = "")
            result <- paste(result, add_values(catsconfig, "suitability", row), sep = "")
        }

        result <- paste(result, "\n")
        name <- catsconfig[row, ]$general.run_name
        name <- make_filename_valid(name)

        for (overlay in overlays) {
            key_name <- sprintf("overlay_%s", overlay)
            result <- paste(result, sprintf("[overlay:%s]", overlay), sep = "\n")
            result <- paste(result, add_values(catsconfig, key_name, row), sep = "\n")
        }


        out_file <- file.path(working_directory, sprintf("catsr_%s.conf", name))
        files <- c(files, out_file)
        sink(out_file)
        cat(result)
        sink()
        cat(sprintf("Wrote configuration file for %s to %s\n", catsconfig[row, ]$general.run_name, out_file))
    }
    return(files)
}
