

library(raster)
library(data.table)
library(plyr)

rotate <- function(x)
    t(apply(x, 2, rev))

post_process <- function(grid, min_prob, multiplier = 1)
{
    # input raster is not necessarily square, so make it square by extendińg

    max_extent         <- max(abs(raster::extent(grid)[]))
    if (max_extent %% 2 == 0) {
        max_extent <- max_extent + 1
    }

    grid <-
        raster::extend(grid,
                       c(-max_extent, max_extent,-max_extent, max_extent),
                       value = 0)

    # counts to probabilities
    grid <- grid / sum(grid[])

    # convert to matrix, and rotate by 90/180/270 degrees
    g    <- raster::as.matrix(grid)

    ext  <- raster::extent(grid)

    g1   <- g
    g2   <- rotate(g1)
    g3   <- rotate(g2)
    g4   <- rotate(g3)

    # average
    g    <- (g1 + g2 + g3 + g4) / 4

    # convert back to raster
    g    <-
        raster::raster(
            vals = g,
            ncols = ncol(grid),
            nrows = nrow(grid),
            resolution = raster::res(grid),
            xmn = ext[1],
            xmx = ext[2],
            ymn = ext[3],
            ymx = ext[4]
        )

    # remove values below threshold
    g[g < min_prob] <- 0

    # cut to size
    g <- raster::trim(g, values = c(0, NA))

    # and apply multiplier
    g  <- g * multiplier

}


#' Create a dispersal raster
#'
#' @param input either a function that takes exactly one integer argument and
#'              returns a vector of that length containing dispersal distance observations
#'              e.g. input(3) -> c(102.3, 24.0, 77.9), or a table (data frame)
#'              with dispersal probabilities. The data frame must have three columns:
#'              uppperBound, lowerBound, density, where upperBound and lowerBound are
#'              the upper and lower dispersal distances, and density is the probability
#'              that a dispersed seed will fall in that range bracket
#' @param outfile the file name of output file, should end in .asc
#' @param resolution the resolution (in m) of the dispersal raster, should match the resolution of your CATS simulation
#' @param repetitions how often should the dispersal process be repeated, default 1000
#' @param events how many dispersal events (per repetition) should be drawn from the provided dispersal distribution, default 1e5
#' @param prob_threshold the minimum dispersal threshold, any values below this threshold will be set to zero, before the raster is trimmed
#' @param multiplier a value between 0 and 1, which will be multiplied with all raster values. Default 1.0. Values less than 1.0
#'        are used to indicate losses in the dispersal process, e.g. to account for seed destruction in the gut
#'
#' @return a raster object: the dispersal raster
#' @export
create_dispersal_raster <- function(input,
                    outfile,
                    resolution,
                    repetitions = 1e3,
                    events = 1e5,
                    prob_threshold = 1e-9,
                    multiplier = 1.0)
{
    # check arguments
    if ((length(repetitions) > 1) | (length(events) > 1)) {
        stop("repetions and event scalars, not vectors")
    }
    if (prob_threshold < 0 | prob_threshold >= 1) {
        stop("minimum probability out of range [0, 1)")
    }

    if (is.data.frame(input) | is.function(input)) {
        #  ok
    } else {
        stop("argument 'input' needs to be a data frame or a function taking one argument")
    }



    initial_grid_length <-
        19          # side length of the output grid raster, odd integer


    if (multiplier <= 0 | multiplier > 1) {
        stop("multiplier out of range (0, 1]")
    }

    if (is.function(input)) {
        # removed

    } else {
        if (is.character(input)) {
            data       <- utils::read.table(input)
        } else {
            data <- input
        }
        if ("density" %in% colnames(data)
            && "upperBound" %in% colnames(data)
            && "lowerBound" %in% colnames(data) )
        {
            # ok
        } else {
            stop("Could not find required columns 'density', 'lowerBound', and 'upperBound' in input")
        }

        dists      <- (data$upperBound + data$lowerBound) / 2
        probs      <- data$density

        probs      <- probs[dists < Inf]
        dists      <- dists[dists < Inf]
    }

    size       <- (resolution * initial_grid_length) / 2
    total      <-
        raster::raster(
            ncols = initial_grid_length,
            nrows = initial_grid_length,
            resolution = c(resolution, resolution),
            vals = 0,
            xmn = -size,
            xmx = size,
            ymn = -size,
            ymx = size
        )

    for (i in 1:repetitions) {
        ptm <- proc.time()
        source_x           <-
            stats::runif(1,
                         min = (-resolution / 2),
                         max = (resolution / 2))
        source_y           <-
            stats::runif(1,
                         min = (-resolution / 2),
                         max = (resolution / 2))
        alpha_hori         <- stats::runif(events, min = 0, max = 360)
        if (is.function(input)) {
            dispersal_distance <- input(events)
        } else {
            dispersal_distance <-
                sample(dists,
                       size = events,
                       prob = probs,
                       replace = T)
        }

        x_coord_seed       <-
            dispersal_distance * cos(alpha_hori * (pi / 180))
        y_coord_seed       <-
            dispersal_distance * sin(alpha_hori * (pi / 180))
        coord_seed         <-
            data.table::data.table(
                x = round((x_coord_seed + source_x) / resolution) * resolution,
                y = round((y_coord_seed + source_y) /
                              resolution) * resolution
            )

        # counts the number of seeds with the same coordinate
        coord_seed         <-
            plyr::count(coord_seed, vars = c("x", "y"))
        # ... and creates a new raster from it
        new_raster         <-
            raster::rasterFromXYZ(coord_seed, res = c(resolution, resolution))

        # pad both rasters to same size
        new_extent         <- raster::extent(new_raster)
        our_extent         <- raster::extent(total)

        # important: remove NAs introduced by rasterFromXYZ
        # total[is.na(total)] <- 0
        new_raster[is.na(new_raster)] <- 0

        total              <-
            raster::extend(total, new_extent, value = 0)
        new_raster         <-
            raster::extend(new_raster, our_extent, value = 0)

        if (raster::extent(total) != raster::extent(new_raster)) {
            print(new_raster)
            print(total)
            stop("mismatched extents!")
        }

        total <- total + new_raster
        # print(proc.time() - ptm)
    }


    total <-
        post_process(total, prob_threshold, multiplier = multiplier)

    raster::writeRaster(total, outfile, format = "ascii", overwrite = TRUE)
    return(total)
}


smaller_grid <-
    function(infile, outfile, threshold, overwrite = FALSE)
    {
        if (threshold <= 0.0 || threshold >= 1.0) {
            stop("threshold out of range (0, 1)")
        }
        grid <- raster::raster(infile)
        old <- c(ncol(grid), nrow(grid), sum(grid[]))
        grid[grid < threshold] <- 0
        grid <- raster::trim(grid, values = c(0, NA))

        new <- c(ncol(grid), nrow(grid), sum(grid[]))


        print(
            sprintf(
                "changed grid from %d x %d cells to %d x %d cells, sum from %f to %f",
                old[1],
                old[2],
                new[1],
                new[2],
                old[3],
                new[3]
            )
        )


        raster::writeRaster(grid, outfile, format = "ascii", overwrite = overwrite)
        return (grid)
    }
