make_filename_valid <- function(name) {
    name <- gsub(" ", "_", name, fixed = T)
    return(name)
}

check_filenames <- function(excerpt, working_directory) {
    valid <- cats_validation_object()
    valid <- add_validation(valid, verify_suitability(excerpt, working_directory))
    return(valid)
}


list_suitability_files <- function(excerpt, working_directory = NULL, check.existence = FALSE) {
    valid_excerpt(excerpt)
    if (check.existence == TRUE && !is.null(working_directory)) {
        if (!dir.exists(working_directory)) {
            stop(sprintf("working directory %s does not exist", working_directory))
        }
    }

    if (is.null(excerpt$suitability.filename_pattern)) {
        return(c())
    }

    if (excerpt$suitability.interpolation_interval > 0 && !grepl("%Y", excerpt$suitability.filename_pattern, fixed = T)) {
        stop("suitability.interpolation_interval greater than 0 requires that config.suitability.filename_pattern contains '%Y', which gets substituted with current simulation year")
    }

    if (is.null(excerpt$general.starting_year) || is.na(excerpt$general.starting_year)) {
        stop("general.starting_year not set")
    }

    if (is.null(excerpt$suitability.interpolation_interval) || is.na(excerpt$suitability.interpolation_interval)) {
        stop("suitability.interpolation_interval not set")
    }


    if (is.null(excerpt$general.time_steps) || is.na(excerpt$general.time_steps)) {
        stop("general.time_steps not set")
    }




    year <- excerpt$general.starting_year
    end_year <- excerpt$general.starting_year + excerpt$general.time_steps
    pattern <- excerpt$suitability.filename_pattern
    step <- excerpt$suitability.interpolation_interval

    years <- c()
    names <- c()
    oks <- c()
    while (year <= end_year) {
        current <- gsub("%Y", year, pattern)

        years <- c(years, year)

        year <- year + step
        if (check.existence == T) {
            ok <- FALSE
            if (R.utils::isAbsolutePath(current) == TRUE && file.exists(current)) {
                ok <- TRUE
            } else if (R.utils::isAbsolutePath(current) == FALSE && file.exists(file.path(working_directory, current))) {
                ok <- TRUE
            }
            oks <- c(oks, ok)
        }

        if (!is.null(working_directory) && !R.utils::isAbsolutePath(current)) {
            current <- file.path(working_directory, current)
        }

        names <- c(names, current)
    }

    if (check.existence) {
        result <- data.frame(year = years, filename = names, exists = oks)
    } else {
        result <- data.frame(year = years, filename = names)
    }




    return(result)
}

list_initial_population_distribution_files <- function(excerpt, working_directory = NULL, check.existence = FALSE) {
    valid_excerpt(excerpt)

    if (check.existence == TRUE && !is.null(working_directory)) {
        if (!dir.exists(working_directory)) {
            stop(sprintf("working directory %s does not exist", working_directory))
        }
    }

    fns <- c(excerpt$species.initial_population_filename)
    if (is.null(excerpt$species.initial_population_filename)) {
        return(data.frame(filenames = c(), exists = c()))
    }



    names <- c()
    oks <- c()

    for (current in fns) {
        if (class(current) != "character") {
            stop("species.initial_population_filename must be  of class character ")
        }

        if (check.existence == T) {
            ok <- FALSE
            if (R.utils::isAbsolutePath(current) == TRUE && file.exists(current)) {
                ok <- TRUE
            } else if (R.utils::isAbsolutePath(current) == FALSE && file.exists(file.path(working_directory, current))) {
                ok <- TRUE
            }
            oks <- c(oks, ok)
        }
        if (!is.null(working_directory) && !R.utils::isAbsolutePath(current)) {
            current <- file.path(working_directory, current)
        }
        names <- c(names, current)
    }

    if (check.existence) {
        result <- data.frame(filename = names, exists = oks)
    } else {
        result <- data.frame(filename = names)
    }




    return(result)
}

list_dispersal_files <- function(excerpt, working_directory = NULL, check.existence = FALSE) {
    valid_excerpt(excerpt)

    if (check.existence == TRUE && !is.null(working_directory)) {
        if (!dir.exists(working_directory)) {
            stop(sprintf("working directory %s does not exist", working_directory))
        }
    }

    fns <- c()
    if (is.null(excerpt$dispersal.kernel_filenames)) {
        return(data.frame(filenames = c(), exists = c()))
    }

    if (class(excerpt$dispersal.kernel_filenames) != "list" && class(excerpt$dispersal.kernel_filenames) != "character") {
        stop("dispersal.kernel_filenames must be a list of character or of class character ")
    }

    if (class(excerpt$dispersal.kernel_filenames) == "list") {
        fns <- unlist(excerpt$dispersal.kernel_filenames)
    } else {
        fns <- c(excerpt$dispersal.kernel_filenames)
    }
    names <- c()
    oks <- c()

    for (current in fns) {
        if (class(current) != "character") {
            stop("dispersal.kernel_filenames must be a list of class character or of class character ")
        }

        if (check.existence == T) {
            ok <- FALSE
            if (R.utils::isAbsolutePath(current) == TRUE && file.exists(current)) {
                ok <- TRUE
            } else if (R.utils::isAbsolutePath(current) == FALSE && file.exists(file.path(working_directory, current))) {
                ok <- TRUE
            }
            oks <- c(oks, ok)
        }
        if (!is.null(working_directory) && !R.utils::isAbsolutePath(current)) {
            current <- file.path(working_directory, current)
        }
        names <- c(names, current)
    }

    if (check.existence) {
        result <- data.frame(filename = names, exists = oks)
    } else {
        result <- data.frame(filename = names)
    }




    return(result)
}

check_all_files <- function(excerpt, working_directory) {
    valid_excerpt(excerpt)
    valid <- cats_validation_object()
    disps <- list_dispersal_files(excerpt, working_directory, TRUE)
    suits <- list_suitability_files(excerpt, working_directory, TRUE)
    inis <- list_initial_population_distribution_files(excerpt, working_directory, TRUE)



    for (row in 1:nrow(suits)) {
        fn <- suits[row, "filename"]
        ex <- suits[row, "exists"]
        if (!ex) {
            valid <- add_error(valid, sprintf("Suitability file %s does not exist", fn))
        }
    }


    for (row in 1:nrow(disps)) {
        fn <- disps[row, "filename"]
        ex <- disps[row, "exists"]
        if (!ex) {
            valid <- add_error(valid, sprintf("Dispersal file %s does not exist", fn))
        }
    }

    for (row in 1:nrow(inis)) {
        fn <- inis[row, "filename"]
        ex <- inis[row, "exists"]
        if (!ex) {
            valid <- add_error(valid, sprintf("Initial population file %s does not exist", fn))
        }
    }
    return(valid)
}
