library(grDevices)
library(proj4)
library(raster)


#' List (hybrid parametrisation mode) configuration field names that can be used with
#' [catsr::create_sensitivity_configs()]
#'
#' @return A vector of configuration parameter names
#' @export
sensitivity_hybrid_fields <- function(){
    fields <- catsr::get_cats_hybrid_fields()
    fields <- fields[fields$required==TRUE, ]
    fields <- fields[grepl("^species.", fields$field), ]
    fields <- fields[fields$field != "species.name" & fields$field != "species.initial_population_filename", ]
    return (fields$field)
}

#' Creates CATS configurations for sensitivity analyses by permutating across all specified values.
#'
#' @param cats_config a data frame containing a CATS configuration, e.g. created with [catsr::cats_default_config()]
#' or loaded from a CATS configuration file with [catsr::load_cats_configuration()].
#' @param ... configuration parameter values that should be varied. See [catsr::sensitivity_hybrid_fields()] for a list of
#' allowed parameters. For each specified configuration parameter a vector of possible values must be specified.
#'
#' @description
#' [catsr::create_sensitivity_configs()] can be used to check the sensitivity of a CATS simulation
#' to the species parametrisation values. Based on the supplied CATS configuration and
#' configuration parameters and their values, a data frame containing all supplied value permutations is returned.
#' The returned data frame can be used [catsr::run_cats_sensitivity()] to run simulations.
#'
#' @details
# 'Currently, this function only supports hybrid mode configurations and fields listed by [catsr::sensitivity_hybrid_fields()].
#' The configuration parameter \code{general.output_interval} will be overridden to the value of \code{general.time_steps},
#' so that only population abundance grids will be generated in the first and last simulation year.
#'
#' @return A data frame containing CATS configurations.
#' @export
create_sensitivity_configs <- function(cats_config, ...)
{
    args <- list(...)


    fields <- sensitivity_hybrid_fields()
    if (is.data.frame(cats_config) == FALSE || nrow(cats_config) != 1) {
        stop("Parameter 'cats_config' must be a data.frame with one row")
    }

    wrong <- c()
    for (field in names(args) ) {

        if (field %in% fields) {
            # ok
        } else {
            wrong <- c(wrong, field)
        }
    }
    if (length(wrong) > 0)  {
         stop(paste("Error: cannot vary fields:", paste(wrong, collapse=", ")))
    }
    if (length(args) == 0) {
        stop("At least one paramter has to be variied")
    }

    permutations <- base::expand.grid(args)
    results <- NULL


    for (row_idx in 1:nrow(permutations)) {
        perm <- permutations[row_idx, , drop=F]
        # check here
        entry <- cats_config
        entry$general.output_interval = entry$general.time_steps

        if (is.null(results)) {
            results <- entry
        } else {
            results <- rbind(results, entry)
        }

        last_row = nrow(results)

        for (col_name in colnames(perm)) {
            results[last_row, col_name] <- perm[1, col_name]

        }




    }
    base::rownames(results) <- NULL
    results$general.run_name <- paste0(results$general.run_name, "-", rownames(results))
    results


}

#' Convenience function to run multiple CATS configurations and visualize ## fixme
#'
#'
#' @param cats_configurations A data.frame containing one or more CATS configurations, e.g. created with [catsr::create_sensitivity_configs()]
#' @param working_directory path to the working directory for the CATS simulation to be run in.
#' All relative paths in the CATS configuration are relative to this directory.
#' @param path_to_cats path to the CATS binary
#' @param output_files optional list of output files returned by [catsr::run_cats_sensitivity()] or [catsr::run_cats()].
#' If provided, simulations are not run, and the visualsation is based directly on this data.
#'
#' @description
#' Thus function will run all provided configurations with CATS and plot the results:
#' The abundance rasters of each individual simulation will be translated to presence/absence data,
#' and summed up over all simulations, showing how the parameter ranges affect distribution ranges.
#'
#' @details
#' Configuration files for the indivudal simulations and their output will be generated in the current directory, and cats-output
#' respectively. A list of generated files will be returned, so further manual analysis is possible.
#'
#' @return A data frame containing the output files of a all started CATS runs,
#' which can be plotted with [catsr::plot_cats_run()] or [catsr::run_cats_sensitivity()]
#' Each row of the data frame corresponds to a single output file. The columns
#' of the data frame are the run name, the replicate number, the year (or NA if not applicable, e.g. for log files),
#' the type of output file("log", "grid-stats", "population") and the file name.
#' 'log' is the command line output of the CATS simulation,
#' 'grid-stats' is a CSV file with summary statistics for a CATS simulation, and
#' 'population' are population abundance files in GeoTIFF format.
#' @export
run_cats_sensitivity <- function(cats_configurations, working_directory, path_to_cats, output_files=NULL)
{
    if (is.null(output_files)) {
    output_files <- catsr::run_cats(cats_configurations, working_directory=working_directory,
                    path_to_cats=path_to_cats, quiet=T)
    }

    cat("Loading and plotting results (last year binary distributions)\n")

    pop_files <- output_files[output_files$year==max(output_files$year, na.rm=T) & output_files$file.type=="population", ]
    rstack <- raster::stack(pop_files$file.name)
    rstack[rstack > 0] <- 1
    rstack <- raster::calc(rstack, sum)
    breaks <- c(0, 0:nrow(pop_files) + 0.5)
    where <- c(0.25, 1:nrow(pop_files))
    labels <- 0:nrow(pop_files)

    raster::plot(rstack,
                 col=rev(grDevices::terrain.colors(nrow(pop_files) + 1)),
                 breaks=breaks,
                 axis.args=list(at=where, labels=labels, cex=3),
                 legend.args=list(text="times occupied", side=4, font=2, line=2.5, cex=2),
                 main=sprintf("Sensitivity analysis:\nNumber of times (out of %d simulations) cells were occupied in year %d",
                              nrow(pop_files),  pop_files$year[1]),
                 cex.main=2, cex.axis=2)

    output_files

}
